import pytest
import os
from unittest import mock

from tfx.components.base.base_component import BaseComponent

from fluent_tfx import PipelineDef, \
    ExampleInputs, SchemaInputs, HyperParameterInputs

from fluent_tfx.pipeline_def import build_step
from tfx.components.base import executor_spec
from tfx.extensions.google_cloud_ai_platform.trainer \
    import executor as ai_platform_trainer_executor


def test_pipeline_def_with_hyperparameters_works(pipeline_def, monkeypatch):
    # Arrange
    mock_with_hyperparameters = mock.Mock()
    monkeypatch.setattr(
        'fluent_tfx.pipeline_def.input_builders.with_hyperparameters', mock_with_hyperparameters)

    # Act
    pipeline_def = pipeline_def.with_hyperparameters('some_uri')
    _, kwargs = mock_with_hyperparameters.call_args

    # Assert
    assert mock_with_hyperparameters.called
    assert kwargs['uri'] == 'some_uri'
    assert pipeline_def.components['user_hyperparameters_importer'] is mock_with_hyperparameters.return_value
    assert pipeline_def.user_hyperparameters_importer is mock_with_hyperparameters.return_value


def test_pipeline_def_tune_works(pipeline_def, monkeypatch):
    # Arrange
    mock_schema = mock.MagicMock()
    pipeline_def.schema_gen = mock_schema

    mock_transform = mock.MagicMock()
    pipeline_def.transform = mock_transform

    mock_module = mock.Mock()

    mock_tuner = mock.Mock()
    monkeypatch.setattr('fluent_tfx.pipeline_def.Tuner', mock_tuner)

    # Act
    pipeline_def = pipeline_def.tune(
        module_file=mock_module, train_args=1, eval_args=2)

    _, kwargs = mock_tuner.call_args

    # Assert
    assert mock_tuner.called
    assert kwargs['module_file'] is mock_module
    assert kwargs['schema'] is SchemaInputs.SCHEMA_CHANNEL(pipeline_def)
    assert kwargs['train_args'] == 1
    assert kwargs['eval_args'] == 2
    assert kwargs['transform_graph'] == mock_transform.outputs['transform_graph']
    assert kwargs['examples'] is ExampleInputs.PREPROCESSED_EXAMPLES(
        pipeline_def)
    assert pipeline_def.components['tuner'] is mock_tuner.return_value
    assert pipeline_def.tuner is mock_tuner.return_value


def test_pipeline_def_tune_example_uses_example_channel_if_specified(pipeline_def, monkeypatch):
    # Arrange
    mock_schema = mock.MagicMock()
    pipeline_def.schema_gen = mock_schema

    mock_transform = mock.MagicMock()
    pipeline_def.transform = mock_transform

    mock_module = mock.Mock()

    example_in = mock.MagicMock()

    def example_in_fn(pipeline_instance):
        return example_in(pipeline_instance)

    mock_tuner = mock.Mock()
    monkeypatch.setattr('fluent_tfx.pipeline_def.Tuner', mock_tuner)

    # Act
    pipeline_def = pipeline_def.tune(
        module_file=mock_module, train_args=1, eval_args=2,
        example_input=example_in_fn)

    _, kwargs = mock_tuner.call_args
    args, _ = example_in.call_args
    # Assert
    assert example_in.called
    assert args[0] is pipeline_def
    assert pipeline_def.cached_example_input is example_in.return_value
    assert pipeline_def.components['tuner'] is mock_tuner.return_value
    assert pipeline_def.tuner is mock_tuner.return_value


def test_pipeline_def_tune_does_not_specify_absent_optional_arguments(pipeline_def, monkeypatch):
    # Arrange
    mock_example_gen = mock.MagicMock()
    pipeline_def.example_gen = mock_example_gen

    mock_schema = mock.MagicMock()
    pipeline_def.schema_gen = mock_schema

    pipeline_def.transform = None

    mock_tuner = mock.Mock()
    monkeypatch.setattr('fluent_tfx.pipeline_def.Tuner', mock_tuner)

    mock_module = mock.Mock()

    # Act
    pipeline_def = pipeline_def.tune(
        module_file=mock_module,
        example_input=ExampleInputs.RAW_EXAMPLES)

    _, kwargs = mock_tuner.call_args

    # Assert
    assert mock_tuner.called
    assert 'train_args' not in kwargs
    assert 'eval_args' not in kwargs
    assert 'transform_graph' not in kwargs
    assert kwargs['examples'] is ExampleInputs.RAW_EXAMPLES(pipeline_def)
    assert pipeline_def.cached_example_input is ExampleInputs.RAW_EXAMPLES(
        pipeline_def)
    assert pipeline_def.components['tuner'] is mock_tuner.return_value
    assert pipeline_def.tuner is mock_tuner.return_value



import pytest
import os
from unittest import mock

from tfx.components.base.base_component import BaseComponent

from fluent_tfx import PipelineDef, \
    ExampleInputs, SchemaInputs, HyperParameterInputs

from fluent_tfx.pipeline_def import build_step
from tfx.components.base import executor_spec
from tfx.extensions.google_cloud_ai_platform.trainer \
    import executor as ai_platform_trainer_executor


def test_pipeline_def_add_custom_component_BaseComponent_works(pipeline_def):
    # Arrange
    fake_components = {'a': 'a_instance', 'b': 'b_instance'}
    pipeline_def.components = fake_components

    mock_base_component = mock.Mock(spec=BaseComponent)

    # Act
    pipeline_def = pipeline_def.add_custom_component(
        name='base_component', component=mock_base_component)

    # Assert
    assert 'base_component' in pipeline_def.components
    assert pipeline_def.components['base_component'] is mock_base_component


def test_pipeline_def_add_custom_component_callable_works(pipeline_def):
    # Arrange
    fake_components = {'a': 'a_instance', 'b': 'b_instance'}
    pipeline_def.components = fake_components

    mock_component_ctor = mock.Mock()

    def callable(components):
        return mock_component_ctor(components)

    # Act
    pipeline_def = pipeline_def.add_custom_component(
        name='some_component', component=callable)

    args, _ = mock_component_ctor.call_args
    existing_components = args[0]

    # Assert
    assert mock_component_ctor.called
    assert existing_components['a'] == 'a_instance'
    assert existing_components['b'] == 'b_instance'

    assert 'some_component' in pipeline_def.components
    assert pipeline_def.components['some_component'] is mock_component_ctor.return_value


def test_pipeline_def_add_custom_component_incompatible_type_raises_value_error(pipeline_def):
    # Act Assert
    with pytest.raises(ValueError):
        pipeline_def.add_custom_component(
            name='asdf_component', component='neither BaseComponent nor callable')



# MIT License

# Copyright(c) 2020 Theodoros Ntakouris

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files(the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and / or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""
This file contains code that the library user would write:

- model generation
- input loading with tensorflow datasets
- hyperparameter search
- serving graph construction and saving

"""
import os
import logging
from typing import Dict, Text, Any, Optional, List

import absl
import fluent_tfx as ftfx
import tensorflow as tf
import kerastuner
from kerastuner.tuners import Hyperband
import tensorflow_transform as tft
import tensorflow_model_analysis as tfma
from tfx.orchestration.beam.beam_dag_runner import BeamDagRunner

from tfx.dsl.component.experimental.decorators import component
from tfx.dsl.component.experimental.annotations import InputArtifact
from tfx.types.standard_artifacts import Model, HyperParameters, Schema

from tfx.components.trainer.fn_args_utils import FnArgs
from tfx.components.trainer.executor import TrainerFnArgs
from tfx.components.tuner.component import TunerFnResult

from tfx.proto import trainer_pb2, evaluator_pb2, pusher_pb2

from google.protobuf import text_format
from tensorflow.python.lib.io import file_io
from tensorflow_metadata.proto.v0 import schema_pb2
from tensorflow_transform.tf_metadata import schema_utils

# pipeline definition is at the bottom of the file

LABEL_KEY = 'lbl'
DENSE_FEATURES = ['a', 'b']
BINARY_FEATURES = ['c']


def _gzip_reader_fn(filenames):
    return tf.data.TFRecordDataset(filenames, compression_type='GZIP')


def _get_serve_tf_examples_fn(model, tf_transform_output):
    model.tft_layer = tf_transform_output.transform_features_layer()

    @tf.function
    def serve_tf_examples_fn(serialized_tf_examples):
        feature_spec = tf_transform_output.raw_feature_spec()
        feature_spec.pop(LABEL_KEY)
        parsed_features = tf.io.parse_example(
            serialized_tf_examples, feature_spec)

        transformed_features = model.tft_layer(parsed_features)

        return model(transformed_features)

    return serve_tf_examples_fn


def preprocessing_fn(inputs: Dict[Text, Any]) -> Dict[Text, Any]:
    outputs = {}
    for feat in DENSE_FEATURES:
        outputs[f'{feat}_xf'] = tft.scale_to_z_score(inputs[feat])

    for feat in BINARY_FEATURES:
        outputs[feat] = inputs[feat]

    outputs[LABEL_KEY] = inputs[LABEL_KEY]
    return outputs


def _input_fn(file_pattern: List[Text],
              tf_transform_output: tft.TFTransformOutput,
              batch_size: int = 32) -> tf.data.Dataset:
    transformed_feature_spec = (
        tf_transform_output.transformed_feature_spec().copy())

    dataset = tf.data.experimental.make_batched_features_dataset(
        file_pattern=file_pattern,
        batch_size=batch_size,
        features=transformed_feature_spec,
        reader=_gzip_reader_fn,
        label_key=LABEL_KEY)

    return dataset


H_SIZE = 'h_size'


def _get_hyperparameters() -> kerastuner.HyperParameters:
    hp = kerastuner.HyperParameters()
    hp.Choice(H_SIZE, [5, 10])
    return hp


def _build_keras_model(hparams: kerastuner.HyperParameters) -> tf.keras.Model:
    features_in = []
    features_in.extend(DENSE_FEATURES)
    features_in.extend(BINARY_FEATURES)

    features_in = [f'{x}_xf' for x in features_in]
    input_layers = {
        colname: tf.keras.layers.Input(
            name=colname, shape=(1,), dtype=tf.float32)
        for colname in features_in
    }

    x = tf.keras.layers.Concatenate(axis=-1)(list(input_layers.values()))

    h = int(hparams.get(H_SIZE))
    x = tf.keras.layers.Dense(
        units=h, activation='relu')(x)

    out = tf.keras.layers.Dense(units=1, activation='sigmoid')(x)

    model = tf.keras.Model(input_layers, out)

    model.compile(
        loss='binary_crossentropy',
        optimizer='adam',
        metrics=[tf.keras.metrics.BinaryAccuracy()])

    model.summary(print_fn=logging.info)
    return model


def tuner_fn(fn_args: FnArgs) -> TunerFnResult:
    train_files = fn_args.train_files
    eval_files = fn_args.eval_files

    tf_transform_output = tft.TFTransformOutput(fn_args.transform_graph_path)

    hparams = _get_hyperparameters()

    tuner = kerastuner.Hyperband(
        hypermodel=_build_keras_model,
        hyperparameters=hparams,
        objective=kerastuner.Objective('binary_accuracy', 'max'),
        factor=3,
        max_epochs=2,
        directory=fn_args.working_dir,
        project_name='ftfx:simple_e2e')

    train_dataset = _input_fn(train_files, tf_transform_output)
    eval_dataset = _input_fn(eval_files, tf_transform_output)

    return TunerFnResult(
        tuner=tuner,
        fit_kwargs={
            'x': train_dataset,
            'validation_data': eval_dataset,
            'steps_per_epoch': fn_args.train_steps,
            'validation_steps': fn_args.eval_steps
        })


def run_fn(fn_args: TrainerFnArgs):
    hparams = fn_args.hyperparameters
    if type(hparams) is dict and 'values' in hparams.keys():
        hparams = hparams['values']

    schema = schema_pb2.Schema()
    schema_text = file_io.read_file_to_string(fn_args.schema_file)
    text_format.Parse(schema_text, schema)
    feature_spec = schema_utils.schema_as_feature_spec(schema).feature_spec

    tf_transform_output = tft.TFTransformOutput(fn_args.transform_output)

    train_dataset = _input_fn(fn_args.train_files, tf_transform_output)
    eval_dataset = _input_fn(fn_args.eval_files, tf_transform_output)

    mirrored_strategy = tf.distribute.MirroredStrategy()
    with mirrored_strategy.scope():
        model = _build_keras_model(hparams=hparams)
    try:
        log_dir = fn_args.model_run_dir
    except KeyError:
        log_dir = os.path.join(os.path.dirname(
            fn_args.serving_model_dir), 'logs')

    # Write logs to path
    tensorboard_callback = tf.keras.callbacks.TensorBoard(
        log_dir=log_dir, update_freq='batch')

    model.fit(
        train_dataset,
        steps_per_epoch=fn_args.train_steps,
        validation_data=eval_dataset,
        validation_steps=fn_args.eval_steps,
        callbacks=[tensorboard_callback])

    signatures = {
        'serving_default':
        _get_serve_tf_examples_fn(model,
                                  tf_transform_output).get_concrete_function(
            tf.TensorSpec(
                shape=[None],
                dtype=tf.string,
                name='examples'))
    }
    model.save(fn_args.serving_model_dir,
               save_format='tf', signatures=signatures)



import pytest
import os
from unittest import mock

from tfx.components.base.base_component import BaseComponent

from fluent_tfx import PipelineDef, \
    ExampleInputs, SchemaInputs, HyperParameterInputs

from fluent_tfx.pipeline_def import build_step
from tfx.components.base import executor_spec
from tfx.extensions.google_cloud_ai_platform.trainer \
    import executor as ai_platform_trainer_executor


def test_evaluate_model_works(pipeline_def, monkeypatch):
    # Arrange
    mock_latest_blessed = mock.MagicMock()
    latest_blessed_ret_val = mock.MagicMock()
    mock_latest_blessed.return_value = latest_blessed_ret_val
    monkeypatch.setattr(
        'fluent_tfx.pipeline_def.input_builders.get_latest_blessed_model_resolver', mock_latest_blessed)

    mock_evaluator = mock.MagicMock()
    mock_evaluator_ret_val = mock.MagicMock()
    mock_evaluator.return_value = mock_evaluator_ret_val
    monkeypatch.setattr('fluent_tfx.pipeline_def.Evaluator', mock_evaluator)

    mock_trainer = mock.MagicMock()
    pipeline_def.trainer = mock_trainer

    mock_raw_examples = mock.MagicMock()
    pipeline_def.example_gen = mock_raw_examples
    pipeline_def.cached_example_input = None

    mock_eval_conf = mock.Mock()

    # Act
    pipeline_def = pipeline_def.evaluate_model(eval_config=mock_eval_conf)
    _, kwargs = mock_evaluator.call_args

    # Assert
    assert mock_evaluator.called
    assert kwargs['eval_config'] is mock_eval_conf
    assert kwargs['model'] is mock_trainer.outputs['model']
    assert kwargs['examples'] is ExampleInputs.RAW_EXAMPLES(pipeline_def)
    assert kwargs['baseline_model'] is latest_blessed_ret_val.outputs['model']

    assert pipeline_def.components['model_evaluator'] is mock_evaluator.return_value
    assert pipeline_def.model_evaluator is mock_evaluator.return_value

    assert pipeline_def.components['latest_blessed_model_resolver'] is mock_latest_blessed.return_value
    assert pipeline_def.latest_blessed_model_resolver is mock_latest_blessed.return_value


def test_pipeline_def_evaluate_model_can_use_custom_example_component(pipeline_def, monkeypatch):
    # Arrange
    mock_latest_blessed = mock.MagicMock()
    latest_blessed_ret_val = mock.MagicMock()
    mock_latest_blessed.return_value = latest_blessed_ret_val
    monkeypatch.setattr(
        'fluent_tfx.pipeline_def.input_builders.get_latest_blessed_model_resolver', mock_latest_blessed)

    mock_evaluator = mock.MagicMock()
    mock_evaluator_ret_val = mock.MagicMock()
    mock_evaluator.return_value = mock_evaluator_ret_val
    monkeypatch.setattr('fluent_tfx.pipeline_def.Evaluator', mock_evaluator)

    mock_trainer = mock.MagicMock()
    pipeline_def.trainer = mock_trainer

    mock_eval_conf = mock.Mock()

    mock_example_provider_component = mock.MagicMock(spec=BaseComponent)

    # Act
    pipeline_def = pipeline_def.evaluate_model(
        eval_config=mock_eval_conf, example_provider_component=mock_example_provider_component)
    _, kwargs = mock_evaluator.call_args

    # Assert
    assert mock_evaluator.called

    assert kwargs['examples'] is mock_example_provider_component.outputs['examples']
    assert pipeline_def.components['model_evaluator_example_provider'] is mock_example_provider_component

    assert pipeline_def.components['model_evaluator'] is mock_evaluator.return_value
    assert pipeline_def.model_evaluator is mock_evaluator.return_value


def test_pipeline_def_evaluate_model_can_use_cached_example_inputs(pipeline_def, monkeypatch):
    # Arrange
    mock_latest_blessed = mock.MagicMock()
    latest_blessed_ret_val = mock.MagicMock()
    mock_latest_blessed.return_value = latest_blessed_ret_val
    monkeypatch.setattr(
        'fluent_tfx.pipeline_def.input_builders.get_latest_blessed_model_resolver', mock_latest_blessed)

    mock_evaluator = mock.MagicMock()
    mock_evaluator_ret_val = mock.MagicMock()
    mock_evaluator.return_value = mock_evaluator_ret_val
    monkeypatch.setattr('fluent_tfx.pipeline_def.Evaluator', mock_evaluator)

    mock_trainer = mock.MagicMock()
    pipeline_def.trainer = mock_trainer

    pipeline_def.example_gen = None

    mock_eval_conf = mock.Mock()

    mock_cached_input = mock.Mock()
    pipeline_def.cached_example_input = mock_cached_input

    # Act
    pipeline_def = pipeline_def.evaluate_model(
        eval_config=mock_eval_conf)
    _, kwargs = mock_evaluator.call_args

    # Assert
    assert mock_evaluator.called

    assert kwargs['examples'] is mock_cached_input
    assert pipeline_def.components['model_evaluator'] is mock_evaluator.return_value
    assert pipeline_def.model_evaluator is mock_evaluator.return_value


def test_pipeline_def_evaluate_model_can_cache_provided_example_input(pipeline_def, monkeypatch):
    # Arrange
    mock_latest_blessed = mock.MagicMock()
    latest_blessed_ret_val = mock.MagicMock()
    mock_latest_blessed.return_value = latest_blessed_ret_val
    monkeypatch.setattr(
        'fluent_tfx.pipeline_def.input_builders.get_latest_blessed_model_resolver', mock_latest_blessed)

    mock_evaluator = mock.MagicMock()
    mock_evaluator_ret_val = mock.MagicMock()
    mock_evaluator.return_value = mock_evaluator_ret_val
    monkeypatch.setattr('fluent_tfx.pipeline_def.Evaluator', mock_evaluator)

    mock_trainer = mock.MagicMock()
    pipeline_def.trainer = mock_trainer

    mock_example_gen = mock.MagicMock()
    pipeline_def.example_gen = mock_example_gen
    pipeline_def.cached_example_input = None

    mock_eval_conf = mock.Mock()

    # Act
    pipeline_def = pipeline_def.evaluate_model(
        eval_config=mock_eval_conf, example_input=ExampleInputs.RAW_EXAMPLES)
    _, kwargs = mock_evaluator.call_args

    # Assert
    assert mock_evaluator.called

    assert kwargs['examples'] is mock_example_gen.outputs['examples']
    assert pipeline_def.cached_example_input is ExampleInputs.RAW_EXAMPLES(
        pipeline_def)
    assert pipeline_def.components['model_evaluator'] is mock_evaluator.return_value
    assert pipeline_def.model_evaluator is mock_evaluator.return_value



# MIT License

# Copyright(c) 2020 Theodoros Ntakouris

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files(the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and / or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import os
from operator import attrgetter

import absl
import fluent_tfx as ftfx
from tfx.orchestration.beam.beam_dag_runner import BeamDagRunner

from tfx.orchestration import metadata
from tfx.types import standard_artifacts


def get_pipeline(pipeline_def: ftfx.PipelineDef) -> ftfx.PipelineDef:
    current_dir = os.path.dirname(
        os.path.realpath(__file__))

    return pipeline_def \
        .from_csv(os.path.join(current_dir, 'data')) \
        .generate_statistics() \
        .infer_schema()


if __name__ == '__main__':
    absl.logging.set_verbosity(absl.logging.ERROR)

    bucket_uri = os.path.join(
        os.path.realpath(__file__), 'bucket')

    pipeline_def = ftfx.PipelineDef(name='schema_generation', bucket=bucket_uri) \
        .with_sqlite_ml_metadata() \

    pipeline_def = get_pipeline(pipeline_def)
    pipeline = pipeline_def.build()

    print(pipeline.components)

    BeamDagRunner().run(pipeline)

    # if you are runnign this first time
    # now, a schema should be saved by default under
    # ./bucket/SchemaGen/schema/3/schema.pbtxt
    # run numbers =
    # 1 = example gen
    # 2 = stasistics
    # 3 = schema gen
    schema_output = os.path.join(
        bucket_uri, 'schema_generation', 'staging', 'SchemaGen', 'schema', '3', 'schema.pbtxt')
    print()
    print(f'Schema at {schema_output}:')
    with open(schema_output, 'r') as f:
        print(f.read())

    # alternatively, you can use our sqlite ml metadata.db to view
    # latest schema artifact directly, by using it's uri
    metadata_connection_config = metadata.sqlite_metadata_connection_config(
        os.path.join(bucket_uri, 'schema_generation', 'metadata.db'))  # defaults to this

    with metadata.Metadata(metadata_connection_config) as store:
        schema_artifacts = store.get_artifacts_by_type(
            standard_artifacts.Schema.TYPE_NAME)

        latest_schema_uri = max(schema_artifacts, key=attrgetter(
            'last_update_time_since_epoch')).uri

        latest_schema_uri = os.path.join(latest_schema_uri, 'schema.pbtxt')

        print(f'Schema at {latest_schema_uri}:')
        with open(latest_schema_uri, 'r') as f:
            print(f.read())



import pytest
from unittest import mock

from fluent_tfx.input_builders import from_csv, from_tfrecord, from_bigquery, \
    with_base_model, with_hyperparameters, with_imported_schema, get_latest_blessed_model_resolver
from tfx.types import standard_artifacts, Channel
from tfx.dsl.experimental.latest_blessed_model_resolver import LatestBlessedModelResolver


def test_from_csv_works(monkeypatch):
    # Arrange
    mock_csv_example_gen = mock.Mock()
    monkeypatch.setattr(
        'fluent_tfx.input_builders.CsvExampleGen', mock_csv_example_gen)

    # Act
    x = from_csv('some_uri', name='some_example',
                 input_config=2, output_config=3)
    _, kwargs = mock_csv_example_gen.call_args

    # Assert
    assert mock_csv_example_gen.called
    assert x is mock_csv_example_gen.return_value
    assert type(kwargs['input_base']) is str
    assert kwargs['input_base'] == 'some_uri'
    assert kwargs['instance_name'] == 'some_example'
    assert kwargs['input_config'] == 2
    assert kwargs['output_config'] == 3


def test_from_csv_does_not_apply_optional_arguments_to_example_gen(monkeypatch):
    # Arrange
    mock_csv_example_gen = mock.Mock()
    monkeypatch.setattr(
        'fluent_tfx.input_builders.CsvExampleGen', mock_csv_example_gen)

    # Act
    x = from_csv('some_uri')
    _, kwargs = mock_csv_example_gen.call_args

    # Assert
    assert mock_csv_example_gen.called
    assert 'instance_name' not in kwargs
    assert 'input_config' not in kwargs
    assert 'output_config' not in kwargs


def test_from_tfrecord_works(monkeypatch):
    # Arrange
    mock_import_example_gen = mock.Mock()
    monkeypatch.setattr(
        'fluent_tfx.input_builders.ImportExampleGen', mock_import_example_gen)

    # Act
    x = from_tfrecord('some_uri', name='some_example',
                      input_config=2, output_config=3)
    _, kwargs = mock_import_example_gen.call_args

    # Assert
    assert mock_import_example_gen.called
    assert x is mock_import_example_gen.return_value
    assert type(kwargs['input_base']) is str
    assert kwargs['input_base'] == 'some_uri'
    assert kwargs['instance_name'] == 'some_example'
    assert kwargs['input_config'] == 2
    assert kwargs['output_config'] == 3


def test_from_tfrecord_does_not_apply_optional_arguments_to_example_gen(monkeypatch):
    # Arrange
    mock_import_example_gen = mock.Mock()
    monkeypatch.setattr(
        'fluent_tfx.input_builders.ImportExampleGen', mock_import_example_gen)

    # Act
    x = from_tfrecord('some_uri')
    _, kwargs = mock_import_example_gen.call_args

    # Assert
    assert mock_import_example_gen.called
    assert 'instance_name' not in kwargs
    assert 'input_config' not in kwargs
    assert 'output_config' not in kwargs


def test_from_bigquery_works(monkeypatch):
    # Arrange
    mock_bigquery_example_gen = mock.Mock()
    monkeypatch.setattr(
        'fluent_tfx.input_builders.BigQueryExampleGen', mock_bigquery_example_gen)

    # Act
    x = from_bigquery('some_query', name='some_example',
                      input_config=2, output_config=3)
    _, kwargs = mock_bigquery_example_gen.call_args

    # Assert
    assert mock_bigquery_example_gen.called
    assert x is mock_bigquery_example_gen.return_value
    assert kwargs['query'] == 'some_query'
    assert kwargs['instance_name'] == 'some_example'
    assert kwargs['input_config'] == 2
    assert kwargs['output_config'] == 3


def test_from_bigquery_does_not_apply_optional_arguments_to_example_gen(monkeypatch):
    # Arrange
    mock_bigquery_example_gen = mock.Mock()
    monkeypatch.setattr(
        'fluent_tfx.input_builders.BigQueryExampleGen', mock_bigquery_example_gen)

    # Act
    x = from_bigquery('query')
    _, kwargs = mock_bigquery_example_gen.call_args

    # Assert
    assert mock_bigquery_example_gen.called
    assert 'instance_name' not in kwargs
    assert 'input_config' not in kwargs
    assert 'output_config' not in kwargs


def test_with_imported_schema_works(monkeypatch):
    # Arrange
    mock_importer_node = mock.Mock()
    monkeypatch.setattr(
        'fluent_tfx.input_builders.ImporterNode', mock_importer_node)

    # Act
    x = with_imported_schema('some_uri')
    _, kwargs = mock_importer_node.call_args

    # Assert
    assert mock_importer_node.called
    assert kwargs['instance_name'] == 'with_imported_schema'
    assert kwargs['source_uri'] == 'some_uri'
    assert kwargs['artifact_type'] is standard_artifacts.Schema


def test_with_base_model_works(monkeypatch):
    # Arrange
    mock_importer_node = mock.Mock()
    monkeypatch.setattr(
        'fluent_tfx.input_builders.ImporterNode', mock_importer_node)

    # Act
    x = with_base_model('some_uri')
    _, kwargs = mock_importer_node.call_args

    # Assert
    assert mock_importer_node.called
    assert kwargs['instance_name'] == 'with_base_model'
    assert kwargs['source_uri'] == 'some_uri'
    assert kwargs['artifact_type'] is standard_artifacts.Model


def test_with_hyperparameters_works(monkeypatch):
    # Arrange
    mock_importer_node = mock.Mock()
    monkeypatch.setattr(
        'fluent_tfx.input_builders.ImporterNode', mock_importer_node)

    # Act
    x = with_hyperparameters('some_uri')
    _, kwargs = mock_importer_node.call_args

    # Assert
    assert mock_importer_node.called
    assert kwargs['instance_name'] == 'with_hyperparameters'
    assert kwargs['source_uri'] == 'some_uri'
    assert kwargs['artifact_type'] is standard_artifacts.HyperParameters


def test_get_latest_blessed_model_resolver_works(monkeypatch):
    # Arrange
    mock_resolver_node = mock.Mock()
    monkeypatch.setattr(
        'fluent_tfx.input_builders.ResolverNode', mock_resolver_node)

    # Act
    x = get_latest_blessed_model_resolver()
    _, kwargs = mock_resolver_node.call_args

    # Assert
    assert mock_resolver_node.called
    assert kwargs['instance_name'] == 'latest_blessed_model_resolver'
    assert kwargs['resolver_class'] is LatestBlessedModelResolver
    assert isinstance(kwargs['model'], Channel)
    assert isinstance(kwargs['model_blessing'], Channel)



import pytest
import os
from unittest import mock

from tfx.components.base.base_component import BaseComponent

from fluent_tfx import PipelineDef, \
    ExampleInputs, SchemaInputs, HyperParameterInputs

from fluent_tfx.pipeline_def import build_step
from tfx.components.base import executor_spec
from tfx.extensions.google_cloud_ai_platform.trainer \
    import executor as ai_platform_trainer_executor


def test_raw_example_inputs_raw_work():
    # Arrange
    mock_pipeline_def = mock.MagicMock()
    mock_pipeline_def.example_gen.outputs = {
        'examples': 'example_channel'}

    # Act
    examples = ExampleInputs.RAW_EXAMPLES(mock_pipeline_def)

    # Assert
    assert examples == 'example_channel'


def test_preprocessed_example_inputs_work():
    # Arrange
    mock_pipeline_def = mock.MagicMock()
    mock_pipeline_def.transform.outputs = {
        'transformed_examples': 'example_channel'}

    # Act
    examples = ExampleInputs.PREPROCESSED_EXAMPLES(mock_pipeline_def)

    # Assert
    assert examples == 'example_channel'


def test_best_hyperparameter_inputs_returns_user_provided_if_no_tuner():
    # Arrange
    mock_pipeline_def = mock.MagicMock()
    mock_pipeline_def.user_hyperparameters_importer.outputs = {
        'result': 'hparams'}

    # Act
    hparams = HyperParameterInputs.BEST_HYPERPARAMETERS(mock_pipeline_def)

    # Assert
    assert hparams == 'hparams'


def test_best_hyperparameter_inputs_returns_user_provided_even_if_tuner_exists():
    # Arrange
    mock_pipeline_def = mock.MagicMock()
    mock_pipeline_def.tuner.outputs = {
        'best_hyperparameters': 'tuner_hparams'}

    mock_pipeline_def.user_hyperparameters_importer.outputs = {
        'result': 'imported_hparams'}

    # Act
    hparams = HyperParameterInputs.BEST_HYPERPARAMETERS(mock_pipeline_def)

    # Assert
    assert hparams == 'imported_hparams'


def test_best_hyperparameter_inputs_returns_tuner_if_not_user_provided():
    # Arrange
    mock_pipeline_def = mock.MagicMock()
    mock_pipeline_def.user_hyperparameters_importer = None
    mock_pipeline_def.tuner.outputs = {
        'best_hyperparameters': 'hparams'}

    # Act
    hparams = HyperParameterInputs.BEST_HYPERPARAMETERS(mock_pipeline_def)

    # Assert
    assert hparams == 'hparams'


def test_schema_channel_inputs_returns_user_provided_if_no_schema_gen():
    # Arrange
    mock_pipeline_def = mock.MagicMock()
    mock_pipeline_def.user_schema_importer.outputs = {
        'result': 'schema'}

    # Act
    hparams = SchemaInputs.SCHEMA_CHANNEL(mock_pipeline_def)

    # Assert
    assert hparams == 'schema'


def test_schema_channel_inputs_returns_user_provided_even_if_schema_gen_exists():
    # Arrange
    mock_pipeline_def = mock.MagicMock()
    mock_pipeline_def.schema_gen.outputs = {
        'schema': 'schema_gen'}
    mock_pipeline_def.user_schema_importer.outputs = {
        'result': 'schema_provided'}

    # Act
    hparams = SchemaInputs.SCHEMA_CHANNEL(mock_pipeline_def)

    # Assert
    assert hparams == 'schema_provided'


def test_schema_channel_inputs_returns_schema_gen_if_not_user_provided():
    # Arrange
    mock_pipeline_def = mock.MagicMock()
    mock_pipeline_def.user_schema_importer = None
    mock_pipeline_def.schema_gen.outputs = {
        'schema': 'schema'}

    # Act
    hparams = SchemaInputs.SCHEMA_CHANNEL(mock_pipeline_def)

    # Assert
    assert hparams == 'schema'



# Lint as: python2, python3
# Copyright 2019 Google LLC. All Rights Reserved.
# Modifications Copyright 2020 Theodoros Ntakouris
# Modified parts: pipeline assembly via fluent_tfx got added
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Python source file include taxi pipeline functions and necesasry utils.

The utilities in this file are used to build a model with native Keras.
This module file will be used in Transform and generic Trainer.

The below example is end to end, provided by google as an example at
https://github.com/tensorflow/tfx/blob/master/tfx/examples/chicago_taxi_pipeline/taxi_utils_native_keras.py

The following file is modified to showcase how to create a sample taxi pipeline with fluent_tfx.
The pipeline definition is located at the bottom
"""

import fluent_tfx as ftfx

from tfx.orchestration.beam.beam_dag_runner import BeamDagRunner

import os
import logging
from typing import List, Text, Dict, Any

import absl
import tensorflow as tf
import tensorflow_transform as tft
import tensorflow_model_analysis as tfma
from tfx.proto import pusher_pb2
from tfx.proto import trainer_pb2
from tfx.components.trainer.executor import TrainerFnArgs

# Categorical features are assumed to each have a maximum value in the dataset.
_MAX_CATEGORICAL_FEATURE_VALUES = [24, 31, 12]

_CATEGORICAL_FEATURE_KEYS = [
    'trip_start_hour', 'trip_start_day', 'trip_start_month',
    'pickup_census_tract', 'dropoff_census_tract', 'pickup_community_area',
    'dropoff_community_area'
]

_DENSE_FLOAT_FEATURE_KEYS = ['trip_miles', 'fare', 'trip_seconds']

# Number of buckets used by tf.transform for encoding each feature.
_FEATURE_BUCKET_COUNT = 10

_BUCKET_FEATURE_KEYS = [
    'pickup_latitude', 'pickup_longitude', 'dropoff_latitude',
    'dropoff_longitude'
]

# Number of vocabulary terms used for encoding VOCAB_FEATURES by tf.transform
_VOCAB_SIZE = 1000

# Count of out-of-vocab buckets in which unrecognized VOCAB_FEATURES are hashed.
_OOV_SIZE = 10

_VOCAB_FEATURE_KEYS = [
    'payment_type',
    'company',
]

# Keys
_LABEL_KEY = 'big_tipper'


def _transformed_name(key):
    return key + '_xf'


def _transformed_names(keys):
    return [_transformed_name(key) for key in keys]


def _gzip_reader_fn(filenames):
    """Small utility returning a record reader that can read gzip'ed files."""
    return tf.data.TFRecordDataset(
        filenames,
        compression_type='GZIP')


def _fill_in_missing(x):
    """Replace missing values in a SparseTensor.

    Fills in missing values of `x` with '' or 0, and converts to a dense tensor.

    Args:
      x: A `SparseTensor` of rank 2.  Its dense shape should have size at most 1
        in the second dimension.

    Returns:
      A rank 1 tensor where missing values of `x` have been filled in.
    """
    if isinstance(x, tf.sparse.SparseTensor):
        default_value = '' if x.dtype == tf.string else 0
        dense_tensor = tf.sparse.to_dense(
            tf.SparseTensor(x.indices, x.values, [x.dense_shape[0], 1]),
            default_value)
    else:
        dense_tensor = x

    return tf.squeeze(dense_tensor, axis=1)


def _get_serve_tf_examples_fn(model, tf_transform_output):
    """Returns a function that parses a serialized tf.Example and applies TFT."""

    model.tft_layer = tf_transform_output.transform_features_layer()

    @tf.function
    def serve_tf_examples_fn(serialized_tf_examples):
        """Returns the output to be used in the serving signature."""
        feature_spec = tf_transform_output.raw_feature_spec()
        feature_spec.pop(_LABEL_KEY)
        parsed_features = tf.io.parse_example(
            serialized_tf_examples, feature_spec)

        transformed_features = model.tft_layer(parsed_features)

        return model(transformed_features)

    return serve_tf_examples_fn


def _input_fn(file_pattern: List[Text],
              tf_transform_output: tft.TFTransformOutput,
              batch_size: int = 200) -> tf.data.Dataset:
    """Generates features and label for tuning/training.

    Args:
      file_pattern: List of paths or patterns of input tfrecord files.
      tf_transform_output: A TFTransformOutput.
      batch_size: representing the number of consecutive elements of returned
        dataset to combine in a single batch

    Returns:
      A dataset that contains (features, indices) tuple where features is a
        dictionary of Tensors, and indices is a single Tensor of label indices.
    """
    transformed_feature_spec = (
        tf_transform_output.transformed_feature_spec().copy())

    dataset = tf.data.experimental.make_batched_features_dataset(
        file_pattern=file_pattern,
        batch_size=batch_size,
        features=transformed_feature_spec,
        reader=_gzip_reader_fn,
        label_key=_transformed_name(_LABEL_KEY))

    return dataset


def _build_keras_model(hidden_units: List[int] = None) -> tf.keras.Model:
    """Creates a DNN Keras model for classifying taxi data.

    Args:
      hidden_units: [int], the layer sizes of the DNN (input layer first).

    Returns:
      A keras Model.
    """
    real_valued_columns = [
        tf.feature_column.numeric_column(key, shape=())
        for key in _transformed_names(_DENSE_FLOAT_FEATURE_KEYS)
    ]
    categorical_columns = [
        tf.feature_column.categorical_column_with_identity(
            key, num_buckets=_VOCAB_SIZE + _OOV_SIZE, default_value=0)
        for key in _transformed_names(_VOCAB_FEATURE_KEYS)
    ]
    categorical_columns += [
        tf.feature_column.categorical_column_with_identity(
            key, num_buckets=_FEATURE_BUCKET_COUNT, default_value=0)
        for key in _transformed_names(_BUCKET_FEATURE_KEYS)
    ]
    categorical_columns += [
        tf.feature_column.categorical_column_with_identity(  # pylint: disable=g-complex-comprehension
            key,
            num_buckets=num_buckets,
            default_value=0) for key, num_buckets in zip(
                _transformed_names(_CATEGORICAL_FEATURE_KEYS),
                _MAX_CATEGORICAL_FEATURE_VALUES)
    ]
    indicator_column = [
        tf.feature_column.indicator_column(categorical_column)
        for categorical_column in categorical_columns
    ]

    model = _wide_and_deep_classifier(
        # TODO(b/139668410) replace with premade wide_and_deep keras model
        wide_columns=indicator_column,
        deep_columns=real_valued_columns,
        dnn_hidden_units=hidden_units or [100, 70, 50, 25])
    return model


def _wide_and_deep_classifier(wide_columns, deep_columns, dnn_hidden_units):
    """Build a simple keras wide and deep model.

    Args:
      wide_columns: Feature columns wrapped in indicator_column for wide (linear)
        part of the model.
      deep_columns: Feature columns for deep part of the model.
      dnn_hidden_units: [int], the layer sizes of the hidden DNN.

    Returns:
      A Wide and Deep Keras model
    """
    # Following values are hard coded for simplicity in this example,
    # However prefarably they should be passsed in as hparams.

    # Keras needs the feature definitions at compile time.
    # TODO(b/139081439): Automate generation of input layers from FeatureColumn.
    input_layers = {
        colname: tf.keras.layers.Input(
            name=colname, shape=(), dtype=tf.float32)
        for colname in _transformed_names(_DENSE_FLOAT_FEATURE_KEYS)
    }
    input_layers.update({
        colname: tf.keras.layers.Input(name=colname, shape=(), dtype='int32')
        for colname in _transformed_names(_VOCAB_FEATURE_KEYS)
    })
    input_layers.update({
        colname: tf.keras.layers.Input(name=colname, shape=(), dtype='int32')
        for colname in _transformed_names(_BUCKET_FEATURE_KEYS)
    })
    input_layers.update({
        colname: tf.keras.layers.Input(name=colname, shape=(), dtype='int32')
        for colname in _transformed_names(_CATEGORICAL_FEATURE_KEYS)
    })

    # TODO(b/161952382): Replace with Keras premade models and
    # Keras preprocessing layers.
    deep = tf.keras.layers.DenseFeatures(deep_columns)(input_layers)
    for numnodes in dnn_hidden_units:
        deep = tf.keras.layers.Dense(numnodes)(deep)
    wide = tf.keras.layers.DenseFeatures(wide_columns)(input_layers)

    output = tf.keras.layers.Dense(
        1, activation='sigmoid')(
            tf.keras.layers.concatenate([deep, wide]))
    output = tf.squeeze(output, -1)

    model = tf.keras.Model(input_layers, output)
    model.compile(
        loss='binary_crossentropy',
        optimizer=tf.keras.optimizers.Adam(lr=0.001),
        metrics=[tf.keras.metrics.BinaryAccuracy()])
    model.summary(print_fn=absl.logging.info)
    return model


# TFX Transform will call this function.
def preprocessing_fn(inputs):
    """tf.transform's callback function for preprocessing inputs.
    Args:
        inputs: map from feature keys to raw not-yet-transformed features.
    Returns:
        Map from string feature key to transformed feature operations.
    """
    outputs = {}
    for key in _DENSE_FLOAT_FEATURE_KEYS:
        # Preserve this feature as a dense float, setting nan's to the mean.
        outputs[_transformed_name(key)] = tft.scale_to_z_score(
            _fill_in_missing(inputs[key]))

    for key in _VOCAB_FEATURE_KEYS:
        # Build a vocabulary for this feature.
        outputs[_transformed_name(key)] = tft.compute_and_apply_vocabulary(
            _fill_in_missing(inputs[key]),
            top_k=_VOCAB_SIZE,
            num_oov_buckets=_OOV_SIZE)

    for key in _BUCKET_FEATURE_KEYS:
        outputs[_transformed_name(key)] = tft.bucketize(
            _fill_in_missing(inputs[key]),
            _FEATURE_BUCKET_COUNT)

    for key in _CATEGORICAL_FEATURE_KEYS:
        outputs[_transformed_name(key)] = _fill_in_missing(inputs[key])

    # TODO(b/157064428): Support label transformation for Keras.
    # Do not apply label transformation as it will result in wrong evaluation.
    outputs[_transformed_name(_LABEL_KEY)] = inputs[_LABEL_KEY]

    return outputs


# TFX Trainer will call this function.
def run_fn(fn_args: TrainerFnArgs):
    """Train the model based on given args.

    Args:
      fn_args: Holds args used to train the model as name/value pairs.
    """
    # Number of nodes in the first layer of the DNN
    first_dnn_layer_size = 100
    num_dnn_layers = 4
    dnn_decay_factor = 0.7

    tf_transform_output = tft.TFTransformOutput(fn_args.transform_output)

    train_dataset = _input_fn(fn_args.train_files, tf_transform_output, 40)
    eval_dataset = _input_fn(fn_args.eval_files, tf_transform_output, 40)

    mirrored_strategy = tf.distribute.MirroredStrategy()
    with mirrored_strategy.scope():
        model = _build_keras_model(
            # Construct layers sizes with exponetial decay
            hidden_units=[
                max(2, int(first_dnn_layer_size * dnn_decay_factor**i))
                for i in range(num_dnn_layers)
            ])

    model.fit(
        train_dataset,
        steps_per_epoch=fn_args.train_steps,
        validation_data=eval_dataset,
        validation_steps=fn_args.eval_steps)

    signatures = {
        'serving_default':
            _get_serve_tf_examples_fn(model,
                                      tf_transform_output).get_concrete_function(
                                          tf.TensorSpec(
                                              shape=[None],
                                              dtype=tf.string,
                                              name='examples')),
    }
    model.save(fn_args.serving_model_dir,
               save_format='tf', signatures=signatures)


def _get_eval_config() -> tfma.EvalConfig:
    return tfma.EvalConfig(
        model_specs=[tfma.ModelSpec(label_key='big_tipper')],
        slicing_specs=[tfma.SlicingSpec()],
        metrics_specs=[
            tfma.MetricsSpec(metrics=[
                tfma.MetricConfig(
                    class_name='BinaryAccuracy',
                    threshold=tfma.MetricThreshold(
                        value_threshold=tfma.GenericValueThreshold(
                            lower_bound={'value': 0.6}),
                        change_threshold=tfma.GenericChangeThreshold(
                            direction=tfma.MetricDirection.HIGHER_IS_BETTER,
                            absolute={'value': -1e-10})))
            ])
        ])


def get_pipeline(pipeline_def: ftfx.pipeline_def):

    _data_root = os.path.join(os.path.dirname(
        __file__), 'data', 'big_tipper_label')
    logging.info(f'Loading csv data from: {_data_root}')

    return pipeline_def \
        .from_csv(uri=_data_root) \
        .generate_statistics() \
        .infer_schema() \
        .preprocess(os.path.realpath(__file__)) \
        .train(os.path.realpath(__file__),
               train_args=trainer_pb2.TrainArgs(num_steps=1000),
               eval_args=trainer_pb2.EvalArgs(num_steps=150)) \
        .evaluate_model(eval_config=_get_eval_config()) \
        .push_to(relative_push_uri='serving_model') \
        .with_beam_pipeline_args([
            '--direct_running_mode=multi_processing',
            '--direct_num_workers=0',
        ])


if __name__ == '__main__':
    absl.logging.set_verbosity(absl.logging.INFO)
    pipeline_def = ftfx.PipelineDef(name='chicago_taxi_pipeline') \
        .with_sqlite_ml_metadata() \
        .cache()

    pipeline_def = get_pipeline(pipeline_def)
    BeamDagRunner().run(pipeline_def.build())



import pytest
import os
from unittest import mock

from tfx.components.base.base_component import BaseComponent

from fluent_tfx import PipelineDef, \
    ExampleInputs, SchemaInputs, HyperParameterInputs

from fluent_tfx.pipeline_def import build_step
from tfx.components.base import executor_spec
from tfx.extensions.google_cloud_ai_platform.trainer \
    import executor as ai_platform_trainer_executor


def test_pipeline_def_generate_statistics_works(pipeline_def, monkeypatch):
    # Arrange
    mock_statistics_gen = mock.Mock()
    monkeypatch.setattr(
        'fluent_tfx.pipeline_def.StatisticsGen', mock_statistics_gen)

    mock_example_gen = mock.MagicMock()
    pipeline_def.example_gen = mock_example_gen

    # Act
    pipeline_def = pipeline_def.generate_statistics()
    _, kwargs = mock_statistics_gen.call_args

    # Assert
    assert mock_statistics_gen.called
    assert kwargs['examples'] is mock_example_gen.outputs['examples']
    assert 'schema' not in kwargs
    assert pipeline_def.components['statistics_gen'] is mock_statistics_gen.return_value
    assert pipeline_def.statistics_gen is mock_statistics_gen.return_value


def test_pipeline_def_generate_statistics_uses_imported_schema_if_specified(pipeline_def, monkeypatch):
    # Arrange
    mock_statistics_gen = mock.Mock()
    monkeypatch.setattr(
        'fluent_tfx.pipeline_def.StatisticsGen', mock_statistics_gen)

    mock_example_gen = mock.MagicMock()
    pipeline_def.example_gen = mock_example_gen

    mock_schema_importer = mock.MagicMock()
    pipeline_def.user_schema_importer = mock_schema_importer

    # Act
    pipeline_def = pipeline_def.generate_statistics()
    _, kwargs = mock_statistics_gen.call_args

    # Assert
    assert mock_statistics_gen.called
    assert 'schema' in kwargs
    assert kwargs['schema'] is mock_schema_importer.outputs['result']
    assert pipeline_def.components['statistics_gen'] is mock_statistics_gen.return_value
    assert pipeline_def.statistics_gen is mock_statistics_gen.return_value


def test_pipeline_def_validate_input_data_works(pipeline_def, monkeypatch):
    # Arrange
    mock_statistics_gen = mock.MagicMock()
    pipeline_def.statistics_gen = mock_statistics_gen
    mock_schema_gen = mock.MagicMock()
    pipeline_def.schema_gen = mock_schema_gen

    mock_example_validator = mock.Mock()
    monkeypatch.setattr(
        'fluent_tfx.pipeline_def.ExampleValidator', mock_example_validator)

    # Act
    pipeline_def = pipeline_def.validate_input_data()
    _, kwargs = mock_example_validator.call_args

    # Assert
    assert mock_example_validator.called
    assert kwargs['statistics'] is mock_statistics_gen.outputs['statistics']
    assert kwargs['schema'] is SchemaInputs.SCHEMA_CHANNEL(pipeline_def)
    assert 'exclude_splits' not in kwargs
    assert pipeline_def.components['example_validator'] is mock_example_validator.return_value
    assert pipeline_def.example_validator is mock_example_validator.return_value


def test_pipeline_def_validate_passes_exclude_splits_arg(pipeline_def, monkeypatch):
    # Arrange
    mock_statistics_gen = mock.MagicMock()
    pipeline_def.statistics_gen = mock_statistics_gen
    mock_schema_gen = mock.MagicMock()
    pipeline_def.schema_gen = mock_schema_gen

    mock_example_validator = mock.Mock()
    monkeypatch.setattr(
        'fluent_tfx.pipeline_def.ExampleValidator', mock_example_validator)

    mock_exclude_split = mock.Mock()

    # Act
    pipeline_def = pipeline_def.validate_input_data(
        exclude_splits=mock_exclude_split)
    _, kwargs = mock_example_validator.call_args

    # Assert
    assert 'exclude_splits' in kwargs
    assert kwargs['exclude_splits'] is mock_exclude_split\



def test_pipeline_def_preprocess_works(pipeline_def, monkeypatch):
    # Arrange
    mock_examples = mock.MagicMock()
    pipeline_def.example_gen = mock_examples

    mock_schema = mock.MagicMock()
    pipeline_def.schema_gen = mock_schema

    mock_transform = mock.Mock()
    monkeypatch.setattr(
        'fluent_tfx.pipeline_def.Transform', mock_transform)

    mock_module = mock.Mock()

    # Act
    pipeline_def = pipeline_def.preprocess(mock_module)
    _, kwargs = mock_transform.call_args

    # Assert
    assert mock_transform.called
    assert kwargs['examples'] is mock_examples.outputs['examples']
    assert kwargs['module_file'] is mock_module
    assert kwargs['schema'] is SchemaInputs.SCHEMA_CHANNEL(pipeline_def)
    assert kwargs['materialize'] == True
    assert pipeline_def.components['transform'] is mock_transform.return_value
    assert pipeline_def.transform is mock_transform.return_value


def test_pipeline_def_preprocess_passes_materialize_arg(pipeline_def, monkeypatch):
    # Arrange
    mock_examples = mock.MagicMock()
    pipeline_def.example_gen = mock_examples

    mock_schema = mock.MagicMock()
    pipeline_def.schema_gen = mock_schema

    mock_transform = mock.Mock()
    monkeypatch.setattr(
        'fluent_tfx.pipeline_def.Transform', mock_transform)

    mock_module = mock.Mock()

    # Act
    pipeline_def = pipeline_def.preprocess(mock_module, materialize=False)
    _, kwargs = mock_transform.call_args

    # Assert
    assert mock_transform.called
    assert kwargs['materialize'] == False


def test_pipeline_def_enable_cache_works(pipeline_def):
    # Arrange
    mock_cache = mock.Mock()

    # Act
    pipeline_def = pipeline_def.cache(mock_cache)

    # Assert
    assert pipeline_def.enable_cache is mock_cache


def test_pipeline_def_with_sqlite_ml_metadata_works(pipeline_def):
    # Act
    pipeline_def = pipeline_def.with_sqlite_ml_metadata()
    metadata_uri = pipeline_def.metadata_connection_config.sqlite.filename_uri

    # Assert
    assert metadata_uri.endswith('metadata.db')
    assert pipeline_def.pipeline_name in metadata_uri
    assert pipeline_def.pipeline_bucket in metadata_uri


def test_pipeline_def_assigns_beam_args(pipeline_def):
    # Arrange
    beam_args = mock.Mock()

    # Act
    pipeline_def = pipeline_def.with_beam_pipeline_args(beam_args)

    # Assert
    assert pipeline_def.beam_pipeline_args is beam_args


def test_pipeline_build_works(pipeline_def, monkeypatch):
    # Arrange
    mock_components = mock.Mock()
    pipeline_def.components = mock_components
    mock_cache = mock.Mock()
    pipeline_def.enable_cache = mock_cache
    mock_metadata_conf = mock.Mock()
    pipeline_def.metadata_connection_config = mock_metadata_conf
    mock_beam_args = mock.Mock()
    pipeline_def.beam_pipeline_args = mock_beam_args

    mock_tfx_pipeline = mock.Mock()
    monkeypatch.setattr('fluent_tfx.pipeline_def.Pipeline', mock_tfx_pipeline)

    # Act
    pipeline = pipeline_def.build()
    _, kwargs = mock_tfx_pipeline.call_args

    # Assert
    assert mock_tfx_pipeline.called
    assert pipeline is mock_tfx_pipeline.return_value

    assert kwargs['components'] is mock_components.values()
    assert kwargs['enable_cache'] is mock_cache

    assert 'metadata_connection_config' in kwargs
    assert kwargs['metadata_connection_config'] is mock_metadata_conf
    assert 'beam_pipeline_args' in kwargs
    assert kwargs['beam_pipeline_args'] is mock_beam_args


def test_pipeline_build_does_not_provide_optional_args_if_not_specified_and_does_not_use_cache_by_default(pipeline_def, monkeypatch):
    # Arrange
    mock_components = mock.Mock()
    pipeline_def.components = mock_components
    pipeline_def.enable_cache = None
    pipeline_def.metadata_connection_config = None
    pipeline_def.beam_pipeline_args = None

    mock_tfx_pipeline = mock.Mock()
    monkeypatch.setattr('fluent_tfx.pipeline_def.Pipeline', mock_tfx_pipeline)

    # Act
    pipeline = pipeline_def.build()
    _, kwargs = mock_tfx_pipeline.call_args

    # Assert
    assert mock_tfx_pipeline.called
    assert pipeline is mock_tfx_pipeline.return_value

    assert kwargs['components'] is mock_components.values()
    assert kwargs['enable_cache'] == False

    assert 'metadata_connection_config' not in kwargs
    assert 'beam_pipeline_args' not in kwargs


def test_build_step_works():
    # Arrange
    class BuildStepTester:
        def __init__(self):
            self.components = {}

        @ build_step('some_component_name')
        def some_build_step(self, ret):
            ret(1)
            return ret

        @ build_step('other_component_name')
        def other_build_step(self, ret):
            ret(2)
            return ret

    bs_tester = BuildStepTester()
    some_comp = mock.Mock()
    other_comp = mock.Mock()

    # Act
    bs_tester.some_build_step(some_comp)
    bs_tester.other_build_step(other_comp)

    # Assert
    assert len(bs_tester.components) == 2
    assert 'some_component_name' in bs_tester.components
    assert bs_tester.components['some_component_name'] is some_comp

    assert 'other_component_name' in bs_tester.components
    assert bs_tester.components['other_component_name'] is other_comp

    some_comp.assert_called_once_with(1)
    other_comp.assert_called_once_with(2)



import os
import logging
import shutil
from typing import Text

import tensorflow as tf

from examples.usage_guide.custom_component_schema import get_pipeline
from fluent_tfx.pipeline_def import PipelineDef
from tfx.orchestration import metadata
from tfx.orchestration.beam.beam_dag_runner import BeamDagRunner
import tfx.components


class CustomComponentSchemaFTFXEndToEndTest(tf.test.TestCase):

    def setUp(self):
        super(CustomComponentSchemaFTFXEndToEndTest, self).setUp()
        self._bucket_dir = os.path.join(os.path.dirname(__file__), 'tmpbucket')
        self._clean_bucket_dir()
        self._initial_pipeline_def = PipelineDef(
            name='custom_component_schema_e2e_test', bucket=self._bucket_dir)

    def _clean_bucket_dir(self):
        if os.path.exists(self._bucket_dir):
            shutil.rmtree(self._bucket_dir)

    def tearDown(self):
        self._clean_bucket_dir()

    def assertComponentExecutedOnce(self, component: Text) -> None:
        component_path = os.path.join(
            self._bucket_dir, self._initial_pipeline_def.pipeline_name, 'staging', component)
        self.assertTrue(tf.io.gfile.exists(component_path))
        outputs = tf.io.gfile.listdir(component_path)
        for output in outputs:
            execution = tf.io.gfile.listdir(
                os.path.join(component_path, output))
            self.assertEqual(1, len(execution))

    def assertPipelineExecution(self) -> None:
        self.assertComponentExecutedOnce('CsvExampleGen')
        self.assertComponentExecutedOnce('StatisticsGen')
        self.assertComponentExecutedOnce('SchemaGen')

    def testCustomComponentSchemaE2EFTFXPipeline(self):
        pipeline_def = get_pipeline(self._initial_pipeline_def)
        pipeline_def = pipeline_def.with_sqlite_ml_metadata().cache()

        expected_execution_count = 5  # 4 components + schema importer
        self.assertLen(pipeline_def.components.values(),
                       expected_execution_count)

        self.assertTrue('example_gen' in pipeline_def.components)
        self.assertIsInstance(
            pipeline_def.components['example_gen'], tfx.components.CsvExampleGen)

        self.assertTrue('statistics_gen' in pipeline_def.components)
        self.assertIsInstance(
            pipeline_def.components['statistics_gen'], tfx.components.StatisticsGen)

        self.assertTrue('schema_gen' in pipeline_def.components)
        self.assertIsInstance(
            pipeline_def.components['schema_gen'], tfx.components.SchemaGen)

        self.assertTrue('user_schema_importer' in pipeline_def.components)
        self.assertIsInstance(
            pipeline_def.components['user_schema_importer'], tfx.components.ImporterNode)

        self.assertTrue('schema_printer' in pipeline_def.components)
        self.assertIsInstance(
            pipeline_def.components['schema_printer'], tfx.components.base.base_component.BaseComponent)

        pipeline = pipeline_def.build()

        BeamDagRunner().run(pipeline)

        artifact_root = os.path.join(
            self._bucket_dir, pipeline_def.pipeline_name)
        self.assertTrue(os.path.join(artifact_root, 'metadata.db'))
        metadata_config = pipeline_def.metadata_connection_config
        with metadata.Metadata(metadata_config) as m:
            artifact_count = len(m.store.get_artifacts())
            execution_count = len(m.store.get_executions())
            self.assertLessEqual(artifact_count, execution_count)
            self.assertEqual(expected_execution_count, execution_count)

        self.assertPipelineExecution()

        BeamDagRunner().run(pipeline)

        # All executions are cached
        with metadata.Metadata(metadata_config) as m:
            artifact_count = len(m.store.get_artifacts())
            self.assertEqual(artifact_count, len(m.store.get_artifacts()))
            self.assertEqual(expected_execution_count * 2,
                             len(m.store.get_executions()))

        BeamDagRunner().run(pipeline)

        # Asserts cache execution.
        with metadata.Metadata(metadata_config) as m:
            # Artifact count is unchanged.
            self.assertEqual(artifact_count, len(m.store.get_artifacts()))
            self.assertEqual(expected_execution_count * 3,
                             len(m.store.get_executions()))


if __name__ == '__main__':
    tf.compat.v1.enable_v2_behavior()
    tf.test.main()



import os
import logging
from typing import Optional, Text, List, Dict, Any, Union
from functools import wraps

import tensorflow_model_analysis as tfma
from tfx.orchestration.pipeline import Pipeline
from tfx.orchestration import metadata, data_types

from tfx.proto import example_gen_pb2, trainer_pb2, \
    infra_validator_pb2, pusher_pb2, bulk_inferrer_pb2
from ml_metadata.proto import metadata_store_pb2

from tfx.dsl.experimental import latest_blessed_model_resolver

from tfx.components.base.base_component import BaseComponent

from tfx.components import CsvExampleGen, ImportExampleGen, \
    StatisticsGen, SchemaGen, Transform, ExampleValidator, ImporterNode, \
    ResolverNode, Trainer, Evaluator, InfraValidator, Pusher, BulkInferrer, Tuner

from tfx.extensions.google_cloud_big_query.example_gen.component import BigQueryExampleGen

from tfx.components.base import executor_spec
from tfx.components.trainer import executor as trainer_executor

from tfx.types import standard_artifacts, Channel
from tfx.extensions.google_cloud_ai_platform.trainer \
    import executor as ai_platform_trainer_executor

import fluent_tfx.input_builders as input_builders


def build_step(component_name: Text):
    """Wraps a function inside a `PipelineDef`.
    Adds the component returned to the dict of used components (`self.components`)
    and then returns self as a fluent builder pattern.

    Args:
        func (function): Function to be wrapped

    Returns: self
    """
    def inner_func(func):
        @wraps(func)
        def wrapper(self, *args, **kwargs):
            component = func(self, *args, **kwargs)
            self.components[component_name] = component
            return self
        return wrapper

    return inner_func


class PipelineDef:
    """The fluent definition class that empowers fluent-tfx.
    Initialise the constructor with a name and use the
    factory methods to construct the pipeline.
    Obtain the tfx pipeline instance by using `.build()`
    """

    def __init__(self, name: Text, bucket: Optional[Text] = None,
                 metadata_connection_config: Optional[metadata_store_pb2.ConnectionConfig] = None):
        """
        Args:
            name (Text): The pipeline's name
            bucket (Text, optional): Intermediate artifacts and staging binaries are going
            to be saved under `{name}/{bucket}/..`. Defaults to './bucket'.
            metadata_connection_config (Optional[metadata_store_pb2.ConnectionConfig], optional):
            Optional ML Metadata configuration for rapid local prototyping you can use `with_sqlite_ml_metadata`. Defaults to None.
        """
        if not bucket:
            bucket = os.path.join(os.getcwd(), 'bucket')

        self.components = {}

        self.pipeline_name = name
        self.pipeline_bucket = bucket
        self.metadata_connection_config = metadata_connection_config
        self.enable_cache = False
        self.beam_pipeline_args = None

        self.cached_example_input = None

        self.example_gen = None
        self.statistics_gen = None
        self.example_validator = None
        self.user_schema_importer = None
        self.schema_gen = None
        self.transform = None
        self.user_hyperparameters_importer = None
        self.trainer = None
        self.tuner = None
        self.infra_validator = None
        self.latest_blessed_model_resolver = None
        self.model_evaluator = None
        self.pusher = None
        self.bulk_inferrer = None

    def with_sqlite_ml_metadata(self):
        """WIll use a sqlite database under {bucket}/{name}/metadata.db as a
        backend for ML Metadata.

        Returns: self
        """
        metadata_connection_string = os.path.join(
            self.pipeline_bucket, self.pipeline_name, 'metadata.db')

        self.metadata_connection_config = metadata. \
            sqlite_metadata_connection_config(metadata_connection_string)

        return self

    def add_custom_component(self, name: Text, component):
        """Adds a custom component given a name to the current components pipeline dict.
        The name should not be one of tensorflow extended's components names, converted to snake_case.

        If the component is not an instance of BaseComponent, it should be a
        callable that expects the self.components dict as input

        Raises:
            ValueError: If provided component is neither a BaseComponent nor a callable

        Returns: self
        """
        if name in self.components:
            logging.info(
                f'{name} is already in the components dict, overriding.')

        if isinstance(component, BaseComponent):
            self.components[name] = component
            return self

        if hasattr(component, '__call__'):
            self.components[name] = component(self.components)
            return self

        raise ValueError(
            f'add_custom_component: Expected callable or BaseComponent, got {component} instead.')
        return self

    @build_step('example_gen')
    def from_csv(self, uri: Text,
                 input_config: Optional[example_gen_pb2.Input] = None,
                 output_config: Optional[example_gen_pb2.Output] = None):
        """Constructs a CsvExampleGen component by using `uri`

        Args:
            uri (Text): Csv file(s) uri
            input_config (Optional[example_gen_pb2.Input], optional): Defaults to None.
            output_config (Optional[example_gen_pb2.Output], optional): Defaults to None.

        Returns: self
        """

        self.example_gen = input_builders.from_csv(
            uri=uri, input_config=input_config, output_config=output_config)
        return self.example_gen

    @build_step('example_gen')
    def from_tfrecord(self, uri: Text,
                      input_config: Optional[example_gen_pb2.Input] = None,
                      output_config: Optional[example_gen_pb2.Output] = None):
        """Constructs an ImportExampleGen component by using `uri`

        Args:
            uri (Text): TFRecord file(s) uri
            input_config (Optional[example_gen_pb2.Input], optional): Defaults to None.
            output_config (Optional[example_gen_pb2.Output], optional): Defaults to None.

        Returns: self
        """
        self.example_gen = input_builders.from_tfrecord(
            uri=uri, input_config=input_config, output_config=output_config)
        return self.example_gen

    @build_step('example_gen')
    def from_bigquery(self, query: Text,
                      input_config: Optional[example_gen_pb2.Input] = None,
                      output_config: Optional[example_gen_pb2.Output] = None):
        """Constructs a BigQueryExampleGen component by using `uri`

        Args:
            query (Text): The query to run
            input_config (Optional[example_gen_pb2.Input], optional): Defaults to None.
            output_config (Optional[example_gen_pb2.Output], optional): Defaults to None.

        Returns: self
        """
        self.example_gen = input_builders.from_bigquery(
            query=query, input_config=input_config, output_config=output_config)
        return self.example_gen

    @build_step('example_gen')
    def from_custom_example_gen_component(self, component: BaseComponent):
        """Uses a custom tfx component to generate example files.
        The component should comply with the naming conventions of the other
        example generating components of tfx. (for example, contain a `.outputs['examples']` attribute)

        Args:
            component (BaseComponent): Your custom, compatible component.

        Returns: self
        """
        self.example_gen = component

        return self.example_gen

    @build_step('user_schema_importer')
    def with_imported_schema(self, uri: Text):
        """Constructs an ImporterNode component that imports
        the schema in the pipelineas an artifact.

        If infer_schema is called, the subsequent components will still use this
        use provided schema, but the SchemaGen component will still produce inferred
        schema artifacts.

        Args:
            uri (Text): Schema .pbtxt file uri

        Returns: self
        """
        self.user_schema_importer = input_builders.with_imported_schema(
            uri=uri)

        return self.user_schema_importer

    @build_step('statistics_gen')
    def generate_statistics(self):
        """Constructs a StatisticsGen component on the example_gen output files

        Returns: self
        """
        args = {
            'examples': self.example_gen.outputs['examples']
        }

        if self.user_schema_importer:
            args['schema'] = self.user_schema_importer.outputs['result']

        self.statistics_gen = StatisticsGen(**args)

        return self.statistics_gen

    @build_step('schema_gen')
    def infer_schema(self, infer_feature_shape: Optional[bool] = False):
        """Constructs a SchemaGen component. a StatisticsGen component via `generate_statistics`
        is required as an input.

        Args:
            infer_feature_shape (bool, optional): Defaults to False.

        Returns: self
        """
        self.schema_gen = SchemaGen(
            statistics=self.statistics_gen.outputs['statistics'],
            infer_feature_shape=infer_feature_shape)

        return self.schema_gen

    @build_step('example_validator')
    def validate_input_data(self, exclude_splits: Optional[List[Text]] = None):
        """Constructs an ExampleValidator component that uses a schema and the output
        of StatisticsGen via `generate_statistics` to validate input data.

        If a user provided schema is specified, it will be used.

        Returns: self
        """
        args = {
            'statistics': self.statistics_gen.outputs['statistics'],
            'schema': SchemaInputs.SCHEMA_CHANNEL(self),
        }

        if exclude_splits:
            args['exclude_splits'] = exclude_splits

        self.example_validator = ExampleValidator(**args)
        return self.example_validator

    @build_step('transform')
    def preprocess(self, module_file: Union[Text, data_types.RuntimeParameter], materialize: bool = True):
        """Constructs a Transform component using examples generated and a schema.

        If a user provided schema is specified, it will be used.

        Args:
            module_file (Union[Text, data_types.RuntimeParameter]): The module file (a/b/c.py)
            that contains the preprocessing_fn function. The signature should be
            `def preprocessing_fn(inputs: Dict[Text, Any]) -> Dict[Text, Any]`

        Returns: self
        """
        args = {
            'examples': self.example_gen.outputs['examples'],
            'module_file': module_file,
            'schema': SchemaInputs.SCHEMA_CHANNEL(self),
            'materialize': materialize
        }

        self.transform = Transform(**args)
        return self.transform

    @build_step('train_base_model')
    def with_base_model(self, uri: Text):
        """Constructs an ImporterNode component that imports a `standard_artifacts.Model`
        artifact to use as a starting point for training.

        Args:
            uri (Text): Model artifact's uri

        Returns: self
        """
        self.train_base_model = input_builders.with_base_model(uri=uri)

        return self.train_base_model

    @build_step('user_hyperparameters_importer')
    def with_hyperparameters(self, uri: Text):
        """Constructs an ImporterNode component that imports a `standard_artifacts.HyperParameters`
        artifact to use for future runs.

        Args:
            uri (Text): Hyperparameter artifact's uri

        Returns: self
        """
        self.user_hyperparameters_importer = input_builders.with_hyperparameters(
            uri=uri)

        return self.user_hyperparameters_importer

    @build_step('tuner')
    def tune(self, module_file:
             Union[Text, data_types.RuntimeParameter],
             train_args: Optional[trainer_pb2.TrainArgs] = None,
             eval_args: Optional[trainer_pb2.EvalArgs] = None,
             example_input=None):
        args = {
            'module_file': module_file,
            'schema': SchemaInputs.SCHEMA_CHANNEL(self)
        }

        if train_args:
            args['train_args'] = train_args

        if eval_args:
            args['eval_args'] = eval_args

        if self.transform:
            args['transform_graph'] = self.transform.outputs['transform_graph']

        if example_input:
            inputs = example_input(self)
            args['examples'] = inputs

            self.cached_example_input = inputs
        else:
            args['examples'] = ExampleInputs.PREPROCESSED_EXAMPLES(
                self)

        self.tuner = Tuner(**args)

        return self.tuner

    @build_step('trainer')
    def train(self, module_file: Union[Text, data_types.RuntimeParameter],
              train_args: Optional[trainer_pb2.TrainArgs] = None,
              eval_args: Optional[trainer_pb2.EvalArgs] = None,
              example_input: Optional = None,
              custom_executor_spec: Optional[executor_spec.ExecutorSpec] = None,
              ai_platform_args: Optional[Dict[Text, Text]] = None,
              custom_config: Optional[Dict[Text, Any]] = None):
        """Constructs a Trainer component given the arguments

        If there is a user provided schema, it will be used.
        Similarly, if there are hyperparameters explicitly defined, they will be used.

        No need to explicitly re-specify example inputs if you've already specified them in `tune()`.

        By default, an trainer_executor.GenericExecutor is used.

        If ai platform training arguments are provided, the executor is set to ai_platform_trainer_executor.GenericExecutor

        Args:
            module_file (Union[Text, data_types.RuntimeParameter]): The module file that contains run_fn.
            The signature of the method must be `def run_fn(fn_args: TrainerFnArgs)`.
            example_input ([type], optional): A `ExampleInputs.{RAW_EXAMPLES, PREPROCESSED_EXAMPLES}` function reference which provides
            the example input channel. Defaults to None, where the input channel is set to the transformed/preprocessed examples.

        Returns: self
        """
        args = {
            'module_file': module_file,
            'schema': SchemaInputs.SCHEMA_CHANNEL(self)
        }

        hparams = HyperParameterInputs.BEST_HYPERPARAMETERS(self)
        if hparams:
            args['hyperparameters'] = hparams

        if not custom_executor_spec:
            custom_executor_spec = executor_spec.ExecutorClassSpec(
                trainer_executor.GenericExecutor)

        args['custom_executor_spec'] = custom_executor_spec

        if custom_config:
            args['custom_config'] = custom_config

        if train_args:
            args['train_args'] = train_args

        if eval_args:
            args['eval_args'] = eval_args

        if self.transform:
            args['transform_graph'] = self.transform.outputs['transform_graph']

        if not example_input and self.cached_example_input:
            args['transformed_examples'] = self.cached_example_input
        elif example_input:
            inputs = example_input(self)
            args['transformed_examples'] = inputs

            self.cached_example_input = self.cached_example_input or inputs
        else:
            args['transformed_examples'] = ExampleInputs.PREPROCESSED_EXAMPLES(
                self)

        if ai_platform_args is not None:
            args.update({
                'custom_executor_spec':
                executor_spec.ExecutorClassSpec(
                    ai_platform_trainer_executor.GenericExecutor
                ),
                    'custom_config': {
                    ai_platform_trainer_executor.TRAINING_ARGS_KEY:
                    ai_platform_args,
                }
            })

        self.trainer = Trainer(**args)
        return self.trainer

    @build_step('model_evaluator')
    def evaluate_model(self, eval_config: tfma.EvalConfig,
                       example_input: Optional = None,
                       example_provider_component: Optional[BaseComponent] = None):
        """Constructs an Evaluator component

        No need to specify example_input if you've already specified it in `train()` or `tune()`.

        Args:
            example_input (Optional, optional): A `ExampleInputs.{RAW_EXAMPLES, PREPROCESSED_EXAMPLES}` function reference which provides
            the example input channel. Defaults to None, where the input channel is set to the transformed/preprocessed examples. Defaults
            to RAW_EXAMPLES.


            example_provider_component (Optional[BaseComponent], optional): An external example input component,
            which should output examples at the `outputs['examples']` attribute.

        Returns: self
        """

        self.latest_blessed_model_resolver = input_builders.get_latest_blessed_model_resolver()
        self.components['latest_blessed_model_resolver'] = self.latest_blessed_model_resolver

        args = {
            'model': self.trainer.outputs['model'],
            'baseline_model': self.latest_blessed_model_resolver.outputs['model'],
            'eval_config': eval_config
        }

        if example_provider_component:
            self.components['model_evaluator_example_provider'] = example_provider_component
            args['examples'] = example_provider_component.outputs['examples']

        if not example_provider_component:
            if not example_input and self.cached_example_input:
                args['examples'] = self.cached_example_input
            elif example_input:
                inputs = example_input(self)
                args['examples'] = inputs

                self.cached_example_input = self.cached_example_input or inputs
            else:
                args['examples'] = ExampleInputs.RAW_EXAMPLES(
                    self)

        self.model_evaluator = Evaluator(**args)

        return self.model_evaluator

    @build_step('infra_validator')
    def infra_validate(self, serving_spec: infra_validator_pb2.ServingSpec,
                       validation_spec: Optional[infra_validator_pb2.ValidationSpec] = None,
                       request_spec: Optional[infra_validator_pb2.RequestSpec] = None,
                       example_input=None):
        """Constructs an InfraValidator component.

        No need to specify example_input if you've already specified it in `train()`, `tune()` or `evaluate_model()`.


        Args:
            example_input ([type], optional): A `ExampleInputs.{RAW_EXAMPLES, PREPROCESSED_EXAMPLES}` function reference which provides
            the example input channel. Defaults to None, where the input channel is set to the transformed/preprocessed examples.

        Returns: self
        """
        args = {
            'model': self.trainer.outputs['model'],
            'serving_spec': serving_spec,
        }

        if validation_spec:
            args['validation_spec'] = validation_spec

        if request_spec:
            args['request_spec'] = request_spec

        if not example_input and self.cached_example_input:
            args['examples'] = self.cached_example_input
        elif example_input:
            inputs = example_input(self)
            args['examples'] = inputs

            # unnecessary to assign `inputs` to `self.cached_example_input` here
            # this is probably the last stop for the cached example input train
        else:
            args['examples'] = ExampleInputs.RAW_EXAMPLES(
                self)

        self.infra_validator = InfraValidator(**args)
        return self.infra_validator

    @build_step('pusher')
    def push_to(self, relative_push_uri: Optional[Text] = None,
                push_destination: Optional[pusher_pb2.PushDestination] = None,
                custom_config: Optional[Dict[Text, Any]] = None,
                custom_executor_spec: Optional[executor_spec.ExecutorSpec] = None):
        """Constructs a Pusher component

        Args:
            relative_push_uri (Optional[Text], optional): The relative to `{bucket}/{name}/` uri to push models to.
            Defaults to None, where a `push_destination` `pusher_pb2` is expected.

        Returns: self
        """
        args = {
            'model': self.trainer.outputs['model'],
        }

        if push_destination:
            args['push_destination'] = push_destination
        elif relative_push_uri:
            args['push_destination'] = pusher_pb2.PushDestination(
                filesystem=pusher_pb2.PushDestination.Filesystem(
                    base_directory=os.path.join(
                        self.pipeline_bucket, self.pipeline_name, relative_push_uri)))

        if self.model_evaluator:
            args['model_blessing'] = self.model_evaluator.outputs['blessing']

        if custom_config:
            args['custom_config'] = custom_config

        if custom_executor_spec:
            args['custom_executor_spec'] = custom_executor_spec

        if self.infra_validator:
            args['infra_blessing'] = self.infra_validator.outputs['blessing']

        self.pusher = Pusher(**args)
        return self.pusher

    @build_step('bulk_inferrer')
    def bulk_infer(self, example_provider_component: BaseComponent,
                   data_spec: Optional[Union[bulk_inferrer_pb2.DataSpec,
                                             Dict[Text, Any]]] = None,
                   model_spec: Optional[Union[bulk_inferrer_pb2.ModelSpec,
                                              Dict[Text, Any]]] = None):
        """Generates a BulkInferrer component that uses trainer model output and model evaluator
        blessing.

        Args:
            example_provider_component (BaseComponent): A user-provided component that provides example tfrecord inputs to the
            `.outputs['examples']` attribute.

        Returns: self
        """
        self.components['bulk_inferrer_example_provider'] = example_provider_component

        args = {
            'examples': example_provider_component.outputs['examples'],
            'model': self.trainer.outputs['model']
        }

        if self.model_evaluator:
            args['model_blessing'] = self.model_evaluator.outputs['blessing']

        if data_spec:
            args['data_spec'] = data_spec

        if model_spec:
            args['model_spec'] = model_spec

        self.bulk_inferrer = BulkInferrer(**args)
        return self.bulk_inferrer

    def cache(self, enable_cache=True):
        """
        Args:
            enable_cache (bool, optional): Defaults to True.

        Returns: self
        """
        self.enable_cache = enable_cache

        return self

    def with_beam_pipeline_args(self, args: Optional[List[Text]]):
        """

        Args:
            args (Optional[List[Text]]): Beam Pipeline Arguments

        Returns: self
        """
        self.beam_pipeline_args = args
        return self

    def build(self) -> Pipeline:
        """Builds the pipeline.

        Returns:
            Pipeline: The native TFX pipeline
        """
        args = {
            'pipeline_name': self.pipeline_name,
            'pipeline_root': os.path.join(self.pipeline_bucket, self.pipeline_name, 'staging'),
            'components': self.components.values(),
            'enable_cache': self.enable_cache or False
        }

        if self.metadata_connection_config:
            args['metadata_connection_config'] = self.metadata_connection_config

        if self.beam_pipeline_args:
            args['beam_pipeline_args'] = self.beam_pipeline_args

        return Pipeline(**args)


class ExampleInputs:
    """Provides accessor functions for channel artifact access, regarding example tf records.
    RAW EXAMPLES -> import example_gen files
    PREPROCESSED_EXAMPLES -> import transformed examples
    """
    @staticmethod
    def _get_raw_examples_channel(pipeline_def: PipelineDef) -> Channel:
        return pipeline_def.example_gen.outputs['examples']

    @staticmethod
    def _get_preprocessed_examples_channel(pipeline_def: PipelineDef) -> Channel:
        return pipeline_def.transform.outputs['transformed_examples']

    RAW_EXAMPLES = _get_raw_examples_channel
    PREPROCESSED_EXAMPLES = _get_preprocessed_examples_channel


class HyperParameterInputs:
    """Provides accessor functions for channel artifact access, regarding hyperparameter artifacts.
    BEST_HYPERPARAMETERS -> if they are specified explicitly, return them. Else, return tuner outputs.
    """
    @staticmethod
    def _get_best_hyperparameters(pipeline_def: PipelineDef):
        if pipeline_def.user_hyperparameters_importer:
            return pipeline_def.user_hyperparameters_importer.outputs['result']

        if pipeline_def.tuner:
            return pipeline_def.tuner.outputs['best_hyperparameters']

        return None

    BEST_HYPERPARAMETERS = _get_best_hyperparameters


class SchemaInputs:
    """Provides accessor functions for channel artifact access, regarding schema artifacts.
    SCHEMA_CHANNEL -> if they schema is specified explicitly, return it. Else, return inferred schema.
    """
    @staticmethod
    def _get_schema_channel(pipeline_def: PipelineDef):
        if pipeline_def.user_schema_importer:
            return pipeline_def.user_schema_importer.outputs['result']

        if pipeline_def.schema_gen:
            return pipeline_def.schema_gen.outputs['schema']

        return None

    SCHEMA_CHANNEL = _get_schema_channel



import pytest
import os
from unittest import mock

from tfx.components.base.base_component import BaseComponent

from fluent_tfx import PipelineDef, \
    ExampleInputs, SchemaInputs, HyperParameterInputs

from fluent_tfx.pipeline_def import build_step
from tfx.components.base import executor_spec
from tfx.extensions.google_cloud_ai_platform.trainer \
    import executor as ai_platform_trainer_executor


def test_pipeline_def_with_imported_schema_works(pipeline_def, monkeypatch):
    # Arrange
    mock_with_imported_schema = mock.Mock()
    monkeypatch.setattr(
        'fluent_tfx.pipeline_def.input_builders.with_imported_schema', mock_with_imported_schema)

    # Act
    pipeline_def = pipeline_def.with_imported_schema(uri='schema_uri')
    _, kwargs = mock_with_imported_schema.call_args

    # Assert
    assert mock_with_imported_schema.called
    assert kwargs['uri'] == 'schema_uri'
    assert pipeline_def.components['user_schema_importer'] is mock_with_imported_schema.return_value
    assert pipeline_def.user_schema_importer is mock_with_imported_schema.return_value


def test_pipeline_def_infer_schema_works(pipeline_def, monkeypatch):
    # Arrange
    mock_statistics_gen = mock.MagicMock()
    pipeline_def.statistics_gen = mock_statistics_gen

    mock_schema_gen = mock.Mock()
    monkeypatch.setattr(
        'fluent_tfx.pipeline_def.SchemaGen', mock_schema_gen)

    mock_infer_feature_shape = mock.Mock()

    # Act
    pipeline_def = pipeline_def.infer_schema(
        infer_feature_shape=mock_infer_feature_shape)
    _, kwargs = mock_schema_gen.call_args

    # Assert
    assert mock_schema_gen.called
    assert kwargs['statistics'] is mock_statistics_gen.outputs['statistics']
    assert kwargs['infer_feature_shape'] is mock_infer_feature_shape
    assert pipeline_def.components['schema_gen'] is mock_schema_gen.return_value
    assert pipeline_def.schema_gen is mock_schema_gen.return_value
